# passerelle.apps.cmis
# Copyright (C) 2016  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from cmislib import CmisClient
from cmislib.model import (PermissionDeniedException, ObjectNotFoundException,
                           UpdateConflictException)
import json
import base64
import logging

from django.db import models
from django.utils.translation import ugettext_lazy as _

from passerelle.base.models import BaseResource
from passerelle.utils.api import endpoint
from passerelle.utils.jsonresponse import APIError


class CMISError(APIError):
    pass


class CmisConnector(BaseResource):
    cmis_endpoint = models.URLField(max_length=400, verbose_name=_('CMIS endpoint'),
                                     help_text=_('URL of the CMIS endpoint'))
    username = models.CharField(max_length=128, verbose_name=_('Service username'))
    password = models.CharField(max_length=128, verbose_name=_('Password'))

    category = _('Business Process Connectors')

    class Meta:
        verbose_name = _('CMIS connector')

    def get_description_fields(self):
        fields = super(CmisConnector, self).get_description_fields()
        return [(x[0], x[1]) for x in fields if x[0].name != 'password']

    @endpoint(methods=['post'], perm='can_upload_file')
    def upload_file(self, request, **kwargs):
        try:
            data = json.loads(request.body)
        except ValueError:
            raise CMISError('invalid JSON string sent.')

        if not set(data.keys()) >= set(['filename', 'path', 'content', 'contentType']):
            missing_keys = ', '.join(set(['filename', 'path', 'content', 'contentType'])
                                     - set(data.keys()))
            raise CMISError('missing keys %s' % missing_keys)
        title = data['filename']
        path = data['path'].encode('utf-8')
        if not path:
            path = '/'
        try:
            content = base64.b64decode(data['content'])
        except (ValueError, TypeError):
            raise CMISError('content must be a base64 string')

        content_type = data['contentType']

        repo_name = data.get('repository_name', '')
        repo = self.cmis_connection_repository(repo_name)
        try:
            folder = repo.getObjectByPath(path)
        except ObjectNotFoundException:
            raise CMISError('path not found on platform.')

        try:
            doc = folder.createDocumentFromString(title, contentString=content, contentType=content_type)
        except UpdateConflictException:
            raise CMISError('the document already exists on platform.')

        return {'data': doc.properties}

    def cmis_connection_repository(self, repo_name):
        try:
            cmis_client = CmisClient(self.cmis_endpoint, self.username, self.password)
        except PermissionDeniedException:
            raise CMISError('wrong username or password to connect to platform.')
        except ObjectNotFoundException:
            raise CMISError('platform endpoint not found.')

        if repo_name:
            for repo in cmis_client.getRepositories():
                if repo_name == repo['repositoryName']:
                    return cmis_client.getRepository(repo['repositoryId'])
            raise CMISError('repository %s not found' % repo_name)
        else:
            return cmis_client.defaultRepository
