# passerelle - uniform access to multiple data sources and services
# Copyright (C) 2017  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import uuid

from django.db import models
from django.template import Context, Template
from django.utils.translation import ugettext_lazy as _

from jsonfield import JSONField

from passerelle.base.models import BaseResource
from passerelle.utils.api import endpoint


def get_hex_uuid():
    return uuid.uuid4().get_hex()


class JsonData(models.Model):
    datastore = models.ForeignKey('JsonDataStore', null=True)

    uuid = models.CharField(_('uuid'), max_length=32,
            default=get_hex_uuid, editable=False, unique=True)
    name_id = models.CharField(max_length=256, blank=True)
    content = JSONField(_('Content'))
    text = models.CharField(max_length=256, blank=True)

    creation_datetime = models.DateTimeField(auto_now_add=True)
    last_update_datetime = models.DateTimeField(auto_now=True)

    def save(self, *args, **kwargs):
        text_value_template = self.datastore.text_value_template
        if text_value_template:
            template = Template(text_value_template)
            context = Context(self.content)
            self.text = template.render(context).strip()
        return super(JsonData, self).save(*args, **kwargs)


class JsonDataStore(BaseResource):
    category = _('Data Sources')
    text_value_template = models.CharField(_('Template for "text" key value'),
            max_length=256, blank=True)

    class Meta:
        verbose_name = _('JSON Data Store')

    @endpoint(perm='can_access', name='data', pattern=r'$',
              description=_('Listing'))
    def list(self, request, name_id=None, **kwargs):
        objects = JsonData.objects.filter(datastore=self)
        if name_id:
            objects = objects.filter(name_id=name_id)
        return {'data': [{'id': x.uuid, 'text': x.text, 'content': x.content} for x in objects]}

    @endpoint(perm='can_access', methods=['post'], name='data',
              pattern=r'create$',
              example_pattern='create',
              description=_('Create'))
    def create(self, request, name_id=None, **kwargs):
        attrs = {
            'content': json.loads(request.body),
            'datastore': self,
        }
        if name_id:
            attrs['name_id'] = name_id
        data = JsonData(**attrs)
        data.save()
        return {'id': data.uuid, 'text': data.text}

    def get_data_object(self, uuid, name_id=None):
        attrs = {
            'uuid': uuid,
            'datastore': self,
        }
        if name_id is not None:
            attrs['name_id'] = name_id
        return JsonData.objects.get(**attrs)

    @endpoint(perm='can_access', methods=['get', 'post'], name='data',
              pattern=r'(?P<uuid>\w+)/$',
              example_pattern='{uuid}/',
              description_get=_('Get'),
              description_post=_('Replace'),
              parameters={'uuid': {'description': _('Object identifier'), 'example_value': '12345'}},
             )
    def get_or_replace(self, request, uuid, name_id=None):
        data = self.get_data_object(uuid, name_id)
        if request.method == 'POST':
            data.content = json.loads(request.body)
            data.save()
        return {'id': data.uuid, 'text': data.text, 'content': data.content}

    @endpoint(perm='can_access', methods=['post'], name='data',
              description=_('Delete'),
              pattern=r'(?P<uuid>\w+)/delete$',
              example_pattern='{uuid}/delete',
              parameters={'uuid': {'description': _('Object identifier'), 'example_value': '12345'}},
              )
    def delete(self, request, uuid, name_id=None):
        self.get_data_object(uuid, name_id).delete()
        return {}
