# passerelle - uniform access to multiple data sources and services
# Copyright (C) 2017  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json

from django.db import models
from django.utils.translation import ugettext_lazy as _

from passerelle.base.models import BaseResource
from passerelle.utils.api import endpoint
from passerelle.utils.jsonresponse import APIError


class Solis(BaseResource):
    service_url = models.URLField(max_length=256, blank=False, verbose_name=_('Service URL'),
                                  help_text=_('Solis API base URL'))
    username = models.CharField(max_length=128, blank=False, verbose_name=_('Username'))
    password = models.CharField(max_length=128, blank=False, verbose_name=_('Password'))

    category = _('Business Process Connectors')

    class Meta:
        verbose_name = _('Solis')

    @endpoint(name='ping', description=_('Check Solis API availability'))
    def ping(self, request):
        message = self.requests.get(
                self.service_url + 'main/isAlive',
                auth=(self.username, self.password)).content
        if message.startswith('Solis API'):
            return {'data': 'pong', 'message': message}
        else:
            raise APIError('message is %s' % message)

    def request(self, endpoint, data=None):
        url = self.service_url + endpoint
        auth = (self.username, self.password)
        headers = {'Accept': 'application/json'}
        if data is None:
            return self.requests.get(url, auth=auth, headers=headers)
        return self.requests.post(url, json=data, auth=auth, headers=headers)

    def get_apa_token(self, user_id, code):
        response = self.request('asg/apa/generationJeton', data={
            'indexIndividu': user_id,
            'codeConfidentiel': code,
        })
        if response.status_code == 403:
            raise APIError('invalid credentials')
        elif response.status_code != 200:
            raise APIError('unknown error status:%s, content:%r' %
                           (response.status_code, response.content))
        return response.json().get('token')

    @endpoint(name='apa-link', methods=['post'], perm='can_access',
              description=_('Create link between name_id and '
                            'Solis APA. payload: name_id, user_id, code'))
    def apa_link(self, request):
        try:
            data = json.loads(request.body)
        except ValueError:
            raise APIError('payload is not a JSON dict')
        if not isinstance(data, dict):
            raise APIError('payload is not a JSON dict')
        if 'name_id' not in data:
            raise APIError('missing name_id')
        if 'user_id' not in data or 'code' not in data:
            raise APIError('missing user_id/code credentials')
        name_id, user_id, code = data['name_id'], data['user_id'], data['code']
        self.get_apa_token(user_id, code)  # invalid credentials raise APIError
        link, created = SolisAPALink.objects.update_or_create(resource=self, name_id=name_id,
                defaults={'user_id': user_id, 'code': code})
        return {'data': {'user_id': user_id,
                         'created': created,
                         'updated': not created}}

    @endpoint(name='apa-unlink', perm='can_access',
              description=_('Delete Solis APA link with name_id'),
              parameters={
                  'name_id': {
                      'description': _('user identifier'),
                      'example_value': '3eb56fc'
                  }
              })
    def apa_unlink(self, request, name_id):
        SolisAPALink.objects.filter(resource=self, name_id=name_id).delete()
        return {'data': {'deleted': True}}

    def apa_infos(self, endpoint, name_id):
        apa_link = SolisAPALink.objects.get(resource=self, name_id=name_id)
        token = self.get_apa_token(apa_link.user_id, apa_link.code)
        endpoint = 'asg/apa/' + endpoint + '/' + token
        response = self.request(endpoint)
        if response.status_code != 200:
            raise APIError('unknown error status:%s, content:%r' %
                           (response.status_code, response.content))
        return {'data': response.json()}

    @endpoint(name='apa-user-info', perm='can_access',
              description=_('Call exportDonneesIndividu on linked Solis APA user'),
              parameters={
                  'name_id': {
                      'description': _('user identifier'),
                      'example_value': '3eb56fc'
                  }
              })
    def apa_user_info(self, request, name_id):
        return self.apa_infos('exportDonneesIndividu', name_id)

    @endpoint(name='apa-user-rights', perm='can_access',
              description=_('Call consultationDeMesDroits on linked Solis APA user'),
              parameters={
                  'name_id': {
                      'description': _('user identifier'),
                      'example_value': '3eb56fc'
                  }
              })
    def apa_user_rights(self, request, name_id):
        return self.apa_infos('consultationDeMesDroits', name_id)

    @endpoint(name='apa-user-request', perm='can_access',
              description=_('Call suiviDemandeEnInstruction on linked Solis APA user'),
              parameters={
                  'name_id': {
                      'description': _('user identifier'),
                      'example_value': '3eb56fc'
                  }
              })
    def apa_user_request(self, request, name_id):
        return self.apa_infos('suiviDemandeEnInstruction', name_id)

    @endpoint(name='apa-user-history', perm='can_access',
              description=_('Call suiviDemandeHistorique on linked Solis APA user'),
              parameters={
                  'name_id': {
                      'description': _('user identifier'),
                      'example_value': '3eb56fc'
                  }
              })
    def apa_user_history(self, request, name_id):
        return self.apa_infos('suiviDemandeHistorique', name_id)


class SolisAPALink(models.Model):
    resource = models.ForeignKey(Solis)
    name_id = models.CharField(blank=False, max_length=256)
    user_id = models.CharField(blank=False, max_length=64)
    code = models.CharField(blank=False, max_length=64)
